/*
 * Daneel - Dalvik to Java bytecode compiler
 * Copyright (C) 2011  IcedRobot team
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * This file is subject to the "Classpath" exception:
 *
 * Linking this library statically or dynamically with other modules is
 * making a combined work based on this library.  Thus, the terms and
 * conditions of the GNU General Public License cover the whole
 * combination.
 *
 * As a special exception, the copyright holders of this library give you
 * permission to link this library with independent modules to produce an
 * executable, regardless of the license terms of these independent
 * modules, and to copy and distribute the resulting executable under terms
 * of your choice, provided that you also meet, for each linked independent
 * module, the terms and conditions of the license of that module.  An
 * independent module is a module which is not derived from or based on
 * this library.  If you modify this library, you may extend this exception
 * to your version of the library, but you are not obligated to do so.  If
 * you do not wish to do so, delete this exception statement from your
 * version.
 */

package org.icedrobot.daneel.dex;

import java.nio.ByteBuffer;

/**
 * A parser class capable of parsing {@code class_def_item} structures as part
 * of DEX files. Keep package-private to hide internal API.
 */
class ClassDef {

    /**
     * Parses a {@code class_def_item} structure in a DEX file at the buffer's
     * current position.
     * 
     * @param buffer The byte buffer to read from.
     * @param dex The DEX file currently being parsed.
     * @return An object representing the parsed data.
     */
    static ClassDef parse(ByteBuffer buffer, DexFile dex) {
        return new ClassDef(buffer, dex);
    }

    private String className;

    private int accessFlags;

    private String superclass;

    private String[] interfaces;

    private String sourceFile;

    private ClassData classData;

    private ClassDef(ByteBuffer buffer, DexFile dex) {
        int classIdx = buffer.getInt();
        accessFlags = buffer.getInt();
        int superclassIdx = buffer.getInt();
        int interfacesOff = buffer.getInt();
        int sourceFileIdx = buffer.getInt();
        int annotationsOff = buffer.getInt();
        int classDataOff = buffer.getInt();
        int staticValuesOff = buffer.getInt();

        // Resolve string and type indices.
        className = dex.getTypeDescriptor(classIdx);
        if (superclassIdx != DexFile.NO_INDEX)
            superclass = dex.getTypeDescriptor(superclassIdx);
        if (sourceFileIdx != DexFile.NO_INDEX)
            sourceFile = dex.getString(sourceFileIdx);

        // Parse associated type_list structure and resolve type indices.
        if (interfacesOff != 0) {
            ByteBuffer buf = dex.getDataBuffer(interfacesOff);
            interfaces = new String[buf.getInt()];
            for (int i = 0; i < interfaces.length; i++)
                interfaces[i] = dex.getTypeDescriptor(buf.getShort());
        }

        // Parse the associated class_data_item structure.
        if (classDataOff != 0) {
            ByteBuffer buf = dex.getDataBuffer(classDataOff);
            classData = ClassData.parse(buf, dex, staticValuesOff);
        }
    }

    public String getClassName() {
        return className;
    }

    public int getAccessFlags() {
        return accessFlags;
    }

    public String getSuperclass() {
        return superclass;
    }

    public String[] getInterfaces() {
        return interfaces;
    }

    public String getSourceFile() {
        return sourceFile;
    }

    public ClassData getClassData() {
        return classData;
    }

    /**
     * Allows the given visitor to visit this class definition.
     * 
     * @param visitor The given DEX class visitor object.
     */
    public void accept(DexClassVisitor visitor) {
        visitor.visit(accessFlags, className, superclass, interfaces);
        if (sourceFile != null)
            visitor.visitSource(sourceFile);
        if (classData != null)
            classData.accept(visitor);
        visitor.visitEnd();
    }
}
