/*
 * Daneel - Dalvik to Java bytecode compiler
 * Copyright (C) 2011  IcedRobot team
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * This file is subject to the "Classpath" exception:
 *
 * Linking this library statically or dynamically with other modules is
 * making a combined work based on this library.  Thus, the terms and
 * conditions of the GNU General Public License cover the whole
 * combination.
 *
 * As a special exception, the copyright holders of this library give you
 * permission to link this library with independent modules to produce an
 * executable, regardless of the license terms of these independent
 * modules, and to copy and distribute the resulting executable under terms
 * of your choice, provided that you also meet, for each linked independent
 * module, the terms and conditions of the license of that module.  An
 * independent module is a module which is not derived from or based on
 * this library.  If you modify this library, you may extend this exception
 * to your version of the library, but you are not obligated to do so.  If
 * you do not wish to do so, delete this exception statement from your
 * version.
 */

package org.icedrobot.daneel.dex.dexlib;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;

import org.icedrobot.daneel.dex.DexClassVisitor;
import org.icedrobot.daneel.dex.DexFieldVisitor;
import org.icedrobot.daneel.dex.DexFileVisitor;
import org.icedrobot.daneel.dex.DexMethodVisitor;
import org.icedrobot.daneel.dex.DexReader;
import org.icedrobot.daneel.dex.Opcode;
import org.jf.dexlib.ClassDataItem;
import org.jf.dexlib.ClassDataItem.EncodedField;
import org.jf.dexlib.ClassDataItem.EncodedMethod;
import org.jf.dexlib.ClassDefItem;
import org.jf.dexlib.CodeItem;
import org.jf.dexlib.DexFile;
import org.jf.dexlib.EncodedArrayItem;
import org.jf.dexlib.FieldIdItem;
import org.jf.dexlib.Item;
import org.jf.dexlib.MethodIdItem;
import org.jf.dexlib.ProtoIdItem;
import org.jf.dexlib.StringIdItem;
import org.jf.dexlib.TypeIdItem;
import org.jf.dexlib.TypeListItem;
import org.jf.dexlib.Code.Instruction;
import org.jf.dexlib.Code.InstructionWithReference;
import org.jf.dexlib.Code.LiteralInstruction;
import org.jf.dexlib.Code.SingleRegisterInstruction;
import org.jf.dexlib.Code.Format.Instruction21c;
import org.jf.dexlib.Code.Format.Instruction21h;
import org.jf.dexlib.Code.Format.Instruction21s;
import org.jf.dexlib.Code.Format.Instruction22c;
import org.jf.dexlib.Code.Format.Instruction31c;
import org.jf.dexlib.Code.Format.Instruction31i;
import org.jf.dexlib.Code.Format.Instruction35c;
import org.jf.dexlib.Code.Format.Instruction51l;
import org.jf.dexlib.EncodedValue.BooleanEncodedValue;
import org.jf.dexlib.EncodedValue.ByteEncodedValue;
import org.jf.dexlib.EncodedValue.CharEncodedValue;
import org.jf.dexlib.EncodedValue.DoubleEncodedValue;
import org.jf.dexlib.EncodedValue.EncodedValue;
import org.jf.dexlib.EncodedValue.FloatEncodedValue;
import org.jf.dexlib.EncodedValue.IntEncodedValue;
import org.jf.dexlib.EncodedValue.LongEncodedValue;
import org.jf.dexlib.EncodedValue.ShortEncodedValue;
import org.objectweb.asm.Type;

public class DexLibDexReader implements DexReader {
    private final DexFile dex;
    private HashMap<String, Integer> classIndexMap;

    public DexLibDexReader(File file) throws IOException {
        dex = new DexFile(file);
    }

    @Override
    public void accept(DexFileVisitor dv) {
        List<ClassDefItem> classItems = dex.ClassDefsSection.getItems();
        for (ClassDefItem classItem : classItems) {
            DexClassVisitor cv = dv.visitClass(classItem.getClassType()
                    .getTypeDescriptor());
            if (cv != null) {
                accept(classItem, cv);
            }
        }
        dv.visitEnd();
    }

    @Override
    public void accept(String className, DexClassVisitor cv)
            throws ClassNotFoundException {
        if (classIndexMap == null) {
            HashMap<String, Integer> map = new HashMap<String, Integer>();
            int index = 0;
            List<ClassDefItem> classItems = dex.ClassDefsSection.getItems();
            for (ClassDefItem classItem : classItems) {
                map.put(asInternalName(classItem.getClassType()).replace('/',
                        '.'), index++);
            }
            classIndexMap = map;
        }

        Integer index = classIndexMap.get(className);
        if (index == null) {
            throw new ClassNotFoundException(className);
        }
        accept(dex.ClassDefsSection.getItemByIndex(index), cv);
    }

    private static void accept(ClassDefItem classItem, DexClassVisitor cv) {
        TypeListItem interfaceListItem = classItem.getInterfaces();
        String[] interfaces = null;
        if (interfaceListItem != null) {
            interfaces = new String[interfaceListItem.getTypeCount()];
            for (int i = 0; i < interfaces.length; i++) {
                interfaces[i] = interfaceListItem.getTypeIdItem(i)
                        .getTypeDescriptor();
            }
        }

        cv.visit(classItem.getAccessFlags(),
                classItem.getClassType().getTypeDescriptor(),
                classItem.getSuperclass().getTypeDescriptor(), interfaces);

        StringIdItem sourceFile = classItem.getSourceFile();
        if (sourceFile != null) {
            cv.visitSource(sourceFile.getStringValue());
        }

        // outer-class

        // annotations

        // attributes

        // inner classes

        // fields
        ClassDataItem classData = classItem.getClassData();
        acceptField(classData.getStaticFields(),
                classItem.getStaticFieldInitializers(), cv);
        acceptField(classData.getInstanceFields(), null, cv);

        // methods
        acceptMethod(classData.getDirectMethods(), cv);
        acceptMethod(classData.getVirtualMethods(), cv);

        cv.visitEnd();
    }

    private static String asInternalName(TypeIdItem typeIdItem) {
        return Type.getType(typeIdItem.getTypeDescriptor()).getInternalName();
    }

    private static void acceptField(EncodedField[] fields,
            EncodedArrayItem staticInitItem, DexClassVisitor cv) {
        EncodedValue[] constants = (staticInitItem == null) ? null
                : staticInitItem.getEncodedArray().values;

        for (int i = 0; i < fields.length; i++) {
            EncodedField field = fields[i];
            FieldIdItem fieldIdItem = field.field;

            DexFieldVisitor fv = cv.visitField(field.accessFlags, fieldIdItem
                    .getFieldName().getStringValue(),
                    fieldIdItem.getFieldType().getTypeDescriptor(),
                    (constants == null ? null : getConstant(constants[i])));
            if (fv != null) {
                // visit annotation

                // visit attribute

                fv.visitEnd();
            }
        }
    }

    private static Object getConstant(EncodedValue value) {
        switch (value.getValueType()) {
        case VALUE_BOOLEAN:
            return ((BooleanEncodedValue) value).value;
        case VALUE_BYTE:
            return ((ByteEncodedValue) value).value;
        case VALUE_SHORT:
            return ((ShortEncodedValue) value).value;
        case VALUE_CHAR:
            return ((CharEncodedValue) value).value;
        case VALUE_INT:
            return ((IntEncodedValue) value).value;
        case VALUE_LONG:
            return ((LongEncodedValue) value).value;
        case VALUE_FLOAT:
            return ((FloatEncodedValue) value).value;
        case VALUE_DOUBLE:
            return ((DoubleEncodedValue) value).value;
        default:
            throw new AssertionError("invalid type " + value.getValueType());
        }
    }

    private static void acceptMethod(EncodedMethod[] methods, DexClassVisitor cv) {
        for (EncodedMethod method : methods) {
            MethodIdItem methodIdItem = method.method;
            ProtoIdItem protoIdItem = methodIdItem.getPrototype();

            DexMethodVisitor mv = cv.visitMethod(method.accessFlags,
                    methodIdItem.getMethodName().getStringValue(), protoIdItem
                            .getConciseIdentity(), protoIdItem.getReturnType()
                            .getTypeDescriptor(),
                    asDescriptorArray(protoIdItem));
            if (mv != null) {
                CodeItem codeItem = method.codeItem;

                // visit annotation default

                // visit annotation

                // visit parameter annotation

                // visit attribute

                mv.visitCode(codeItem.getRegisterCount(), -1, -1); // FIXME
                                                                   // insSize/outSize
                acceptCode(codeItem, mv);
                mv.visitEnd();
            }
        }
    }

    private static String[] asDescriptorArray(ProtoIdItem protoIdItem) {
        TypeListItem parameters = protoIdItem.getParameters();
        if (parameters == null) {
            return new String[0];
        }
        int length = parameters.getTypeCount();
        String[] types = new String[length];
        for (int i = 0; i < length; i++) {
            types[i] = parameters.getTypeIdItem(i).getTypeDescriptor();
        }
        return types;
    }

    // opcode categories
    private static final int INSTR = 0;
    private static final int INSTR_OP = 1;
    private static final int INSTR_UNARY_OP = 2;
    private static final int INSTR_BIN_OP = 3;
    private static final int INSTR_BIN_AND_LITERAL = 4;
    private static final int INSTR_CONST_STRING = 5;
    private static final int INSTR_CONST_U32 = 6;
    private static final int INSTR_CONST_U64 = 7;
    private static final int INSTR_CLASS = 8;
    private static final int INSTR_INSTANCEOF = 9;
    private static final int INSTR_GOTO = 10;
    private static final int INSTR_IF_TESTZ = 11;
    private static final int INSTR_IF_TEST = 12;
    private static final int INSTR_SWITCH = 13;
    private static final int INSTR_ARRAY = 14;
    private static final int INSTR_FIELD = 15;
    private static final int INSTR_METHOD = 16;
    private static final int INSTR_FILLED_NEW_ARRAY = 17;
    private static final int INSTR_NEW_ARRAY = 18;
    private static final int INSTR_FILL_ARRAY_DATA = 19;

    private static final int[] OPCODE_CATEGORY;
    static { // see OpcodesEncoder
        int[] categories = new int[256];
        String text = "ACCCCCCCCCBBBBABBBGGGGHHHHFFIBBI"
                + "JCISRRTBKKKNNDDDDDMMMMMMLLLLLL@@"
                + "@@@@OOOOOOOOOOOOOOPPPPPPPPPPPPPP"
                + "PPPPPPPPPPPPPPQQQQQ@QQQQQ@@CCCCC"
                + "CCCCCCCCC@CCCCCCDDDDDDDDDDDDDDDD"
                + "DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD"
                + "DDDDDDDDDDDDDDDDEEEEEEEEEEEEEEEE"
                + "EEE@@@@@@@@@@@@@@@@@@@@@@@@@@@@@";
        for (int i = 0; i < text.length(); i++) {
            categories[i] = text.charAt(i) - 'A';
        }
        OPCODE_CATEGORY = categories;
    }

    private static void acceptCode(CodeItem codeItem, DexMethodVisitor mv) {
        Instruction[] instrs = codeItem.getInstructions();
        for (int i = 0; i < instrs.length; i++) {
            Instruction instr = instrs[i];
            acceptInstr(instr, mv);
        }
    }

    private static void acceptInstr(Instruction instr, DexMethodVisitor mv) {
        org.jf.dexlib.Code.Opcode dexOpcode = instr.opcode;
        Opcode opcode = Opcode.getOpcode(dexOpcode.value);
        switch (OPCODE_CATEGORY[dexOpcode.value]) {

        case INSTR:
            mv.visitInstr(opcode);
            break;

        case INSTR_FIELD: {
            FieldIdItem field;
            int vsrcOrDest;
            int vref;
            if (instr instanceof Instruction21c) {
                Instruction21c i21c = (Instruction21c) instr;
                field = (FieldIdItem) i21c.getReferencedItem();
                vsrcOrDest = i21c.getRegisterA();
                vref = 0; // no instance
            } else {
                Instruction22c i22c = (Instruction22c) instr;
                field = (FieldIdItem) i22c.getReferencedItem();
                vsrcOrDest = i22c.getRegisterA();
                vref = i22c.getRegisterB();
            }
            mv.visitInstrField(opcode, vsrcOrDest, vref, field
                    .getContainingClass().getTypeDescriptor(), field.getFieldName()
                    .getStringValue(), field.getFieldType().getTypeDescriptor());
            break;
        }

        case INSTR_CONST_U64: {
            int vdest = ((SingleRegisterInstruction) instr).getRegisterA();
            long value = ((LiteralInstruction) instr).getLiteral();
            mv.visitInstrConstU64(opcode, vdest, value);
            break;
        }

        case INSTR_CONST_STRING: {
            Item<?> item = ((InstructionWithReference) instr)
                    .getReferencedItem();
            int registerA = ((SingleRegisterInstruction) instr).getRegisterA();
            mv.visitInstrConstString(opcode, registerA,
                    ((StringIdItem) item).getStringValue());
            break;
        }

        case INSTR_METHOD: { // TODO implement range !!!
            Instruction35c i35c = (Instruction35c) instr;
            MethodIdItem methodIdItem = (MethodIdItem) i35c.getReferencedItem();
            String desc = methodIdItem.getPrototype().getPrototypeString();
            int vpacked = i35c.getRegisterG() << 12 | i35c.getRegisterF() << 8
                    | i35c.getRegisterE() << 4 | i35c.getRegisterD();
            mv.visitInstrMethod(opcode, i35c.getRegCount(),
                    i35c.getRegisterA(), vpacked,
                    methodIdItem.getContainingClass().getTypeDescriptor(),
                    methodIdItem.getMethodName().getStringValue(), desc);
            break;
        }

        default:
            throw new AssertionError("unknown category for " + dexOpcode);
        }
    }
}
