/*
 * Daneel - Dalvik to Java bytecode compiler
 * Copyright (C) 2011  IcedRobot team
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * This file is subject to the "Classpath" exception:
 *
 * Linking this library statically or dynamically with other modules is
 * making a combined work based on this library.  Thus, the terms and
 * conditions of the GNU General Public License cover the whole
 * combination.
 *
 * As a special exception, the copyright holders of this library give you
 * permission to link this library with independent modules to produce an
 * executable, regardless of the license terms of these independent
 * modules, and to copy and distribute the resulting executable under terms
 * of your choice, provided that you also meet, for each linked independent
 * module, the terms and conditions of the license of that module.  An
 * independent module is a module which is not derived from or based on
 * this library.  If you modify this library, you may extend this exception
 * to your version of the library, but you are not obligated to do so.  If
 * you do not wish to do so, delete this exception statement from your
 * version.
 */

package org.icedrobot.daneel.loader;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.LinkedList;
import java.util.List;
import java.util.zip.ZipEntry;

import org.icedrobot.daneel.DaneelException;
import org.icedrobot.daneel.dex.DexFile;
import org.icedrobot.daneel.rewriter.DexRewriter;

/**
 * A class loader capable of loading classes compiled for the Dalvik VM. The
 * classes are loaded by transforming them into the Java VM class file format
 * and then defining them in the host VM itself.
 */
public class DaneelClassLoader extends ClassLoader {

    /** The list of all files containing classes. */
    private final DexFile[] classFiles;

    /** The list of all files containing resources. */
    private final ApkFile[] resourceFiles;

    /**
     * Constructs a new class loader. Keep a constructor with such a signature
     * around so that this class loader can be used as a system class loader.
     * For details take a look at the {@link ClassLoader#getSystemClassLoader()}
     * documentation.
     * 
     * @param parent The parent class loader for delegation.
     */
    public DaneelClassLoader(ClassLoader parent) {
        this(parent, defaultFiles());
    }

    /**
     * Constructs a new class loader for a given set of files from which to load
     * classes and resources.
     * 
     * @param parent The parent class loader for delegation.
     * @param files The set of files from which to load.
     */
    public DaneelClassLoader(ClassLoader parent, File[] files) {
        super(parent);

        // Split the given files into APK and DEX files.
        List<DexFile> dexs = new LinkedList<DexFile>();
        List<ApkFile> apks = new LinkedList<ApkFile>();
        for (File file : files) {
            try {
                if (file.getName().endsWith(".apk")) {
                    ApkFile apk = new ApkFile(file);
                    dexs.add(apk.getClasses());
                    apks.add(apk);
                } else {
                    DexFile dex = DexFile.parse(file);
                    dexs.add(dex);
                }
            } catch (IOException e) {
                throw new DaneelException("Unable to setup class loader.", e);
            }
        }

        classFiles = dexs.toArray(new DexFile[dexs.size()]);
        resourceFiles = apks.toArray(new ApkFile[apks.size()]);
    }

    /**
     * Method which tries to find a class for the given name.
     */
    @Override
    protected Class<?> findClass(String name) throws ClassNotFoundException {
        System.out.printf("Trying to find class '%s' ...\n", name);

        // Iterate over all files containing classes.
        for (DexFile dex : classFiles) {
            try {
                byte[] bytecode = DexRewriter.rewrite(name, dex);
                return defineClass(name, bytecode, 0, bytecode.length);
            } catch (ClassNotFoundException e) {
                // Ignore and skip to the next file.
            }
        }

        // Unable to find class definition for given class name.
        throw new ClassNotFoundException();
    }

    /**
     * Method which tries to find a resource for the given name.
     */
    @Override
    protected URL findResource(String name) {
        System.out.printf("Trying to find resource '%s' ...\n", name);

        // Iterate over all files containing resources.
        for (ApkFile apk : resourceFiles) {
            ZipEntry entry = apk.getEntry(name);
            if (entry != null)
                return apk.getJarURL(entry);
        }

        // Unable to find resource for given resource name.
        return null;
    }

    /**
     * Helper method to get the default set of files to load from.
     */
    private static File[] defaultFiles() {
        String path = System.getProperty("daneel.class.path");
        if (path == null || path.isEmpty())
            return new File[0];
        String[] paths = path.split(File.pathSeparator);
        File[] files = new File[paths.length];
        for (int i = 0; i < paths.length; i++)
            files[i] = new File(paths[i]);
        return files;
    }
}
