/*
 * Daneel - Dalvik to Java bytecode compiler
 * Copyright (C) 2011  IcedRobot team
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * This file is subject to the "Classpath" exception:
 *
 * Linking this library statically or dynamically with other modules is
 * making a combined work based on this library.  Thus, the terms and
 * conditions of the GNU General Public License cover the whole
 * combination.
 *
 * As a special exception, the copyright holders of this library give you
 * permission to link this library with independent modules to produce an
 * executable, regardless of the license terms of these independent
 * modules, and to copy and distribute the resulting executable under terms
 * of your choice, provided that you also meet, for each linked independent
 * module, the terms and conditions of the license of that module.  An
 * independent module is a module which is not derived from or based on
 * this library.  If you modify this library, you may extend this exception
 * to your version of the library, but you are not obligated to do so.  If
 * you do not wish to do so, delete this exception statement from your
 * version.
 */

package org.icedrobot.daneel.rewriter;

import static org.objectweb.asm.Opcodes.*;

import static org.icedrobot.daneel.rewriter.Register.*;
import static org.icedrobot.daneel.rewriter.Registers.*;

import java.util.Map;

import org.icedrobot.daneel.dex.DexClassVisitor;
import org.icedrobot.daneel.dex.DexFieldVisitor;
import org.icedrobot.daneel.dex.DexMethodVisitor;
import org.icedrobot.daneel.dex.DexReader;
import org.icedrobot.daneel.dex.Label;
import org.icedrobot.daneel.dex.Opcode;
import org.icedrobot.daneel.util.TypeUtil;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.FieldVisitor;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Type;
import org.objectweb.asm.tree.AbstractInsnNode;
import org.objectweb.asm.tree.LdcInsnNode;
import org.objectweb.asm.tree.VarInsnNode;

public class DexRewriter implements DexClassVisitor {
    private final ClassVisitor cv;

    /**
     * Rewrites a class from Dalvik bytecode into Java bytecode representation.
     * The source class is given by specifying its name and the DEX source it is
     * originating from.
     * 
     * @param name The name of the class to rewrite as a normal fully qualified
     *        class name.
     * @param reader The reader for the DEX source to read from.
     * @return The Java bytecode for the given class.
     * @throws ClassNotFoundException In case the source does not define a class
     *         by that name.
     */
    public static byte[] rewrite(String name, DexReader reader)
            throws ClassNotFoundException {
        ClassWriter writer = new ClassWriter(ClassWriter.COMPUTE_MAXS);
        reader.accept("L" + name.replace('.', '/') + ";", new DexRewriter(writer));
        return writer.toByteArray();
    }

    public DexRewriter(ClassVisitor cv) {
        this.cv = cv;
    }

    @Override
    public void visit(int access, String name, String supername,
            String[] interfaces) {
        name = TypeUtil.convertDescToInternal(name);
        supername = TypeUtil.convertDescToInternal(supername);
        interfaces = (interfaces == null) ? null : TypeUtil
                .convertDescToInternals(interfaces);
        cv.visit(V1_6, access, name, null, supername, interfaces);
    }

    @Override
    public void visitSource(String source) {
        cv.visitSource(source, null);
    }

    @Override
    public DexFieldVisitor visitField(int access, String name, String type,
            Object value) {
        final FieldVisitor fv = cv.visitField(access, name, type, null, value);
        if (fv == null) {
            return null;
        }

        return new DexFieldVisitor() {
            @Override
            public void visitEnd() {
                fv.visitEnd();
            }
        };
    }

    @Override
    public DexMethodVisitor visitMethod(int access, String name, String shorty,
            String returnType, String[] parameterTypes) {
        String desc = TypeUtil.convertProtoToDesc(returnType, parameterTypes);
        MethodVisitor mv = cv.visitMethod(access, name, desc, null, null);
        if (mv == null) {
            return null;
        }

        return new MethodRewriter(mv);
    }

    @Override
    public void visitEnd() {
        cv.visitEnd();
    }

    static class MethodRewriter implements DexMethodVisitor {
        final PatchMethodVisitor mv;

        private Interpreter interpreter;
        private int returnRegisterType; // type of the register used to stored
                                        // the return value

        public MethodRewriter(MethodVisitor mv) {
            this.mv = new PatchMethodVisitor(mv);
        }

        private static int getTypeFromASMType(Type type) {
            switch (type.getSort()) {
            case Type.BOOLEAN:
                return BOOLEAN_TYPE;
            case Type.BYTE:
                return BYTE_TYPE;
            case Type.CHAR:
                return CHAR_TYPE;
            case Type.SHORT:
                return SHORT_TYPE;
            case Type.INT:
                return INT_TYPE;
            case Type.LONG:
                return LONG_TYPE;
            case Type.FLOAT:
                return FLOAT_TYPE;
            case Type.DOUBLE:
                return DOUBLE_TYPE;
            case Type.VOID:
                return VOID_TYPE;
            default:
                return OBJECT_TYPE;
            }
        }

        private static int getTypeFromFieldDescriptor(String descriptor) {
            return getTypeFromASMType(Type.getType(descriptor));
        }

        private static int getReturnTypeFromMethodDescriptor(String descriptor) {
            return getTypeFromASMType(Type.getReturnType(descriptor));
        }

        private static void throwNewAssertionError(Opcode opcode) {
            throw new AssertionError("invalid opcode " + opcode);
        }

        @Override
        public void visitCode(int registerSize, int insSize, int outSize) {
            mv.visitCode();
            interpreter = new Interpreter(registerSize);
        }

        @Override
        public void visitInstr(Opcode opcode) {
            if (opcode == Opcode.RETURN_VOID) {
                mv.visitInsn(RETURN);
                interpreter.setDead();
                return;
            }
            if (opcode == Opcode.NOP) {
                mv.visitInsn(NOP);
                return;
            }
            throwNewAssertionError(opcode);
        }

        @Override
        public void visitInstrField(Opcode opcode, int vsrcOrDest, int vref,
                String owner, String name, String desc) {
            owner = TypeUtil.convertDescToInternal(owner);
            if (opcode.compareTo(Opcode.IGET) >= 0
                    && opcode.compareTo(Opcode.IGET_SHORT) <= 0) {
                mv.visitVarInsn(ALOAD, vref);
                mv.visitFieldInsn(GETFIELD, owner, name, desc);
                int type = getTypeFromFieldDescriptor(desc);
                interpreter.store(vsrcOrDest, type);
                mv.visitVarInsn(Register.getJavaOpcode(type, ISTORE),
                        vsrcOrDest);
            } else if (opcode.compareTo(Opcode.SGET) >= 0
                    && opcode.compareTo(Opcode.SGET_SHORT) <= 0) {
                mv.visitFieldInsn(GETSTATIC, owner, name, desc);
                int type = getTypeFromFieldDescriptor(desc);
                interpreter.store(vsrcOrDest, type);
                mv.visitVarInsn(Register.getJavaOpcode(type, ISTORE),
                        vsrcOrDest);
            } else if (opcode.compareTo(Opcode.IPUT) >= 0
                    && opcode.compareTo(Opcode.IPUT_SHORT) <= 0) {
                mv.visitVarInsn(ALOAD, vref);
                int type = getTypeFromFieldDescriptor(desc);
                mv.visitVarInsn(Register.getJavaOpcode(type, ILOAD), vsrcOrDest);
                mv.visitFieldInsn(PUTFIELD, owner, name, desc);
                interpreter.load(vsrcOrDest, type);
            } else if (opcode.compareTo(Opcode.SPUT) >= 0
                    && opcode.compareTo(Opcode.SPUT_SHORT) <= 0) {
                int type = getTypeFromFieldDescriptor(desc);
                mv.visitVarInsn(Register.getJavaOpcode(type, ILOAD), vsrcOrDest);
                mv.visitFieldInsn(PUTSTATIC, owner, name, desc);
                interpreter.load(vsrcOrDest, type);
            } else {
                throwNewAssertionError(opcode);
            }
        }

        @Override
        public void visitInstrConstString(Opcode opcode, int vdest, String value) {
            mv.visitLdcInsn(value);
            mv.visitVarInsn(ASTORE, vdest);
            interpreter.store(vdest, OBJECT_TYPE);
        }

        @Override
        public void visitInstrMethod(Opcode opcode, int num, int va,
                int vpacked, String owner, String name, String desc) {
            owner = TypeUtil.convertDescToInternal(owner);
            switch (opcode) {
            case INVOKE_VIRTUAL:
            case INVOKE_STATIC:
            case INVOKE_SUPER:
            case INVOKE_DIRECT:
                int javaOpcode;
                int[] registers = { getRegisterD(vpacked),
                        getRegisterE(vpacked), getRegisterF(vpacked),
                        getRegisterG(vpacked), va };
                int r;
                if (opcode == Opcode.INVOKE_STATIC) {
                    r = 0;
                    javaOpcode = INVOKESTATIC;
                } else {
                    r = 1;
                    javaOpcode = (opcode == Opcode.INVOKE_SUPER) ? INVOKESPECIAL
                            : (opcode == Opcode.INVOKE_DIRECT && "<init>"
                                    .equals(name)) ? INVOKESPECIAL
                                    : INVOKEVIRTUAL;
                    int register = registers[0];
                    mv.visitVarInsn(ALOAD, register);
                    interpreter.load(register, OBJECT_TYPE);
                }

                Type[] types = Type.getArgumentTypes(desc);
                for (int i = 0; i < types.length; i++) {
                    Type asmType = types[i];
                    int type = getTypeFromASMType(asmType);
                    int register = registers[r];
                    mv.visitVarInsn(Register.getJavaOpcode(type, ILOAD),
                            register);
                    interpreter.load(register, type);
                    r += asmType.getSize();
                }
                mv.visitMethodInsn(javaOpcode, owner, name, desc);
                returnRegisterType = getReturnTypeFromMethodDescriptor(desc);
                return;
            default:
                throwNewAssertionError(opcode);
            }
        }

        @Override
        public void visitInstrConstU64(Opcode opcode, final int vdest,
                final long value) {
            mv.setPatchMode();
            mv.visitLdcInsn(value);
            final AbstractInsnNode ldc = mv.getLastInsnNode();
            mv.visitVarInsn(LSTORE, vdest);
            final AbstractInsnNode lstore = mv.getLastInsnNode();
            interpreter.storeArbitraryType(vdest, U64_TYPE, new Patchable() {
                @Override
                protected void patch(int registerType) {
                    if (registerType == DOUBLE_TYPE) {
                        mv.patch(ldc,
                                new LdcInsnNode(Double.longBitsToDouble(value)));
                        mv.patch(lstore, new VarInsnNode(DSTORE, vdest));
                    }
                }
            });
        }

        // unimplemented
        @Override
        public void visitLabel(Label label) {
            throw new UnsupportedOperationException("NYI");
        }

        @Override
        public void visitInstrArray(Opcode opcode, int vsrcOrDest, int varray,
                int vindex) {
            throw new UnsupportedOperationException("NYI");
        }

        @Override
        public void visitInstrBinOp(Opcode opcode, int vdest, int vsrc,
                int vsrc2) {
            throw new UnsupportedOperationException("NYI");
        }

        @Override
        public void visitInstrBinOpAndLiteral(Opcode opcode, int vdest,
                int vsrc, int value) {
            throw new UnsupportedOperationException("NYI");
        }

        @Override
        public void visitInstrClass(Opcode opcode, int vdest, String type) {
            throw new UnsupportedOperationException("NYI");
        }

        @Override
        public void visitInstrConstU32(Opcode opcode, int vdest, int value) {
            throw new UnsupportedOperationException("NYI");
        }

        @Override
        public void visitInstrFillArrayData(Opcode opcode, int vsrc,
                Object arrayData) {
            throw new UnsupportedOperationException("NYI");
        }

        @Override
        public void visitInstrFilledNewArray(Opcode opcode, int num, int va,
                int vpacked, String type) {
            throw new UnsupportedOperationException("NYI");
        }

        @Override
        public void visitInstrGoto(Opcode opcode, Label label) {
            throw new UnsupportedOperationException("NYI");
        }

        @Override
        public void visitInstrIfTest(Opcode opcode, int vsrc1, int vsrc2,
                Label label) {
            throw new UnsupportedOperationException("NYI");
        }

        @Override
        public void visitInstrIfTestZ(Opcode opcode, int vsrc, Label label) {
            throw new UnsupportedOperationException("NYI");
        }

        @Override
        public void visitInstrInstanceof(Opcode opcode, int vdest, int vsrc,
                String type) {
            throw new UnsupportedOperationException("NYI");
        }

        @Override
        public void visitInstrNewArray(Opcode opcode, int vdest, int vsize,
                String type) {
            throw new UnsupportedOperationException("NYI");
        }

        @Override
        public void visitInstrOp(Opcode opcode, int srcOrDst) {
            throw new UnsupportedOperationException("NYI");
        }

        @Override
        public void visitInstrSwitch(Opcode opcode, int vsrc,
                Map<Integer, Label> labels) {
            throw new UnsupportedOperationException("NYI");
        }

        @Override
        public void visitInstrUnaryOp(Opcode opcode, int vdest, int vsrc) {
            throw new UnsupportedOperationException("NYI");
        }

        @Override
        public void visitEnd() {
            mv.visitMaxs(-1, -1);
            mv.visitEnd();
        }

        @Override
        public void visitTryCatch(Label start, Label end, Label handler,
                String type) {
            // TODO Auto-generated method stub

        }
    }
}
