/*
 * Daneel - Dalvik to Java bytecode compiler
 * Copyright (C) 2011  IcedRobot team
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * This file is subject to the "Classpath" exception:
 *
 * Linking this library statically or dynamically with other modules is
 * making a combined work based on this library.  Thus, the terms and
 * conditions of the GNU General Public License cover the whole
 * combination.
 *
 * As a special exception, the copyright holders of this library give you
 * permission to link this library with independent modules to produce an
 * executable, regardless of the license terms of these independent
 * modules, and to copy and distribute the resulting executable under terms
 * of your choice, provided that you also meet, for each linked independent
 * module, the terms and conditions of the license of that module.  An
 * independent module is a module which is not derived from or based on
 * this library.  If you modify this library, you may extend this exception
 * to your version of the library, but you are not obligated to do so.  If
 * you do not wish to do so, delete this exception statement from your
 * version.
 */

package org.icedrobot.daneel.rewriter;

import org.objectweb.asm.Opcodes;

public class Register {
    // the order is the same order as java bytecode IALOAD/IASTORE order (see
    // #getJavaOpcode)
    public static final int INT_TYPE = 0; // signed int
    public static final int LONG_TYPE = 1; // signed long
    public static final int FLOAT_TYPE = 2; // 32 bits float
    public static final int DOUBLE_TYPE = 3; // 64 bits float
    public static final int OBJECT_TYPE = 4; // object reference
    public static final int BYTE_TYPE = 5; // byte or boolean
    public static final int CHAR_TYPE = 6; // char
    public static final int SHORT_TYPE = 7; // short
    public static final int BOOLEAN_TYPE = 8; // boolean
    public static final int VOID_TYPE = 9; // void, added to simplify management

    public static final int U32_TYPE = -1; // arbitrary 32bits constant
    public static final int U64_TYPE = -2; // arbitrary 64bits constant

    public static final int NO_TYPE = -3; // unassigned register

    private static final int ARBITRARY_MASK = Integer.MIN_VALUE;

    private int type;

    private Patchable patchable;

    Register() {
        type = NO_TYPE;
        // and patchable == null
    }

    public void load(int expectedType) {
        assert (expectedType & ARBITRARY_MASK) == 0;

        if ((this.type & ARBITRARY_MASK) != 0 && patchable != null) {
            patchable.doPatch(expectedType);
            patchable = null;
        }
        this.type = expectedType;
    }

    public void store(int type) {
        this.type = type;
        this.patchable = null;
    }

    public void storeArbitraryType(int arbitraryType, Patchable patchable) {
        this.type = arbitraryType;
        this.patchable = patchable;
    }

    public void copy(Register register) {
        this.type = register.type;
        this.patchable = register.patchable;
    }

    public void merge(Register register) {
        int thisType = this.type;
        int registerType = register.type;
        if (thisType == registerType) { // short cut
            return;
        }

        boolean thisIsArbitrary = (thisType & ARBITRARY_MASK) != 0;
        boolean registerIsArbitrary = (registerType & ARBITRARY_MASK) != 0;

        if (thisIsArbitrary) {
            if (registerIsArbitrary) {
                // do nothing, if they are not compatible, the VM verifier will
                // emit a VerifyError
                return;
            }
            if (patchable != null) { // otherwise, it's an uninitialized value
                patchable.doPatch(registerType);
                patchable = null;
            }
            this.type = registerType;
            return;
        }

        // if register.type is not an arbitrary, this means that the code may be
        // not verifiable
        // it's ok if the register is no more used
        if (!registerIsArbitrary) {
            return;
        }

        if (register.patchable != null) {
            register.patchable.doPatch(thisType);
            register.patchable = null;
        }
        register.type = thisType;
    }

    public static int getJavaOpcode(int type, int opcode) {
        assert (type & ARBITRARY_MASK) == 0;

        if (opcode == Opcodes.IALOAD || opcode == Opcodes.IASTORE) {
            return opcode + ((type != BOOLEAN_TYPE) ? type : BYTE_TYPE);
        }

        return opcode + ((type > 4) ? 0 : type); // type > 4, integer numeric
                                                 // promotion
    }

    @Override
    public String toString() {
        switch (type) {
        case OBJECT_TYPE:
            return "object";
        case INT_TYPE:
            return "int";
        case FLOAT_TYPE:
            return "float";
        case LONG_TYPE:
            return "long";
        case DOUBLE_TYPE:
            return "double";
        case U32_TYPE:
            return "u32";
        case U64_TYPE:
            return "u64";
        case NO_TYPE:
            return "uninitialized";
        }
        throw new AssertionError();
    }
}
